/*******************************************************************************
	DIRECT TO PLATE OR FILM AND IMAGE OUTPUT INTERFACE
	==================================================

	THE SOFTWARE IS PROVIDED \"AS-IS\" AND WITHOUT WARRANTY OF ANY KIND,
	EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
	WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

	IN NO EVENT SHALL PETER KOLLER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
	INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
	THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT 
	OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

	(c) COPYRIGHT 2022 PETER KOLLER.
*******************************************************************************/
/*******************************************************************************
	Output functions for rasteriser data
	Any device wanting to perform direct to film or direct to plate, or
	output to am image must implement the functions behind the headers
	defined here
*******************************************************************************/
#ifndef     MAUL_API_RASTEROUTPUT
#define     MAUL_API_RASTEROUTPUT

#define METADATA_TAGFILE _T("metadata.tag")
/* The metadata tag file consists of the following lines:
<Filename>				Original document file name and path
PPI xppi, yppi			x and y resolution in pixels per inch of the original document
FIRST <LEFT/RIGHT>		FIRST LEFT or FIRST RIGHT is used to determine which page is on the front and which on the back of the printed result
CSV:
<filename>.csv			One line for each csv file defining the separations
*/

/* Available interface types */
#define INTERFACE_TYPE_IMAGE			0x01
#define INTERFACE_TYPE_DIRECT			0x02

/* For INTERFACE_TYPE_IMAGE */
#define INTERFACE_TOPDOWNIMG			0x01 /* BMP is a top down Image */
#define INTERFACE_BOTTOMUPIMG			0x02 /* TIFF is a bottom up image */

/* For INTERFACE_TYPE_DIRECT */
#define INTERFACE_PLATEPARALLEL			0x00 /* All pages across the plate are opened, and output line for line */
#define INTERFACE_PAGEFORPAGE			0x01 /* Each page is opened in turn, and output line for line */
#define INTERFACE_METADATAXML			0x02 /* The output is an xml file containing file names and page positions, this is mostly used for testing */
#define INTERFACE_AUTOPLATEFEED			0x10 /* The device is capable of printing several plates using autofeed to load each plate */

/* Helper to read the 1 of n interface types */
#define InterfaceType(it) ((it) & 0x0f)

/* NOTES:
INTERFACE_PLATEPARALLEL mode: All pages must have the same pixel resolution. This is because the data is output line for line, and each line must correspond to the same distance.
							  The interface must support opening simultaneous raster handles for all pages to be output.
							  Output is from left-top to right-bottom of the entire plate. Raster images (pages) cannot be rotated 90, but only 180
							  Open(...) does not need to return a file path.

INTERFACE_PAGEFORPAGE mode:   Pages can have different pixel resolutions.
							  The DTF/DPT device must be capable of moving back to the left-top corner coordinates of each page to be output.
							  Output is from left-top to right-bottom of each page. Raster images (pages) cannot be rotated 90, but only 180
							  Open(...) does not need to return a file path.

INTERFACE_METADATAXML mode:   The pixel resolution of pages is not checked.
							  Open(...) will open a file according to METACONTROL, and data is written to this file as xml.

INTERFACE_AUTOPLATEFEED:	  If set, Open() is called without Close() to feed a new plate. The driver closes the previous plate, loads a new plate, and reopens the new plate.
 */

 /* Available position flags for SetPosition() below */
#define SETPOSN_ENDOFLINE	0x01 /* For INTERFACE_PLATEPARALLEL: Move down one pixel line (and to the start on the left margin), the newPosition field will be NULL */
#define SETPOSN_MOVETO		0x02 /* For INTERFACE_PLATEPARALLEL: The device may only be able to move down..., the newPosition field will be from left and top margins in 0.01 mm increments  */

 /* Some predefined error flags for SetPosition() below */
#define SETPOSN_SUCCESS		0x00 /* The new position was set successfully */
#define SETPOSNERR_INVALID	0x01 /* Cannot move in that direction */
#define SETPOSNERR_RANGE	0x02 /* The new corrdinates are out of range */
#define SETPOSNERR_HANDLE	0x03 /* Invalid handle */

#ifdef _WIN32		//if compiling for windows

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*******************************************************************************
	Items that would already be defined if the Maul Publisher API were included
	This allows this header to stand alone if the remainder is not used.
*******************************************************************************/
#ifndef MAUL_DLL_INTERFACE
typedef unsigned APIRET;
#ifndef _inline
#define	_inline				inline
#endif
/* This structure is used to verify that objects are navigated correctly.*/
typedef struct			_strcident
{
	USHORT		ident;		//structure identifier
	USHORT		cb;			//size of structure before obj data
}	STRCIDL, *PSTRCIDL;
#ifndef RENDERDATA_DEFINED
#define RENDERDATA_DEFINED
#define SZ_RENDERDATANAME		80
	/* Note that the screen and ink id values are used to sort which ink comes first. */
	typedef struct _renderData
	{
		ULONG	cb;				/* The size of this structure. */
		ULONG	flags;			/* Reserved - do not use. */
		SHORT	angle;			/* The rasterisation angle 0 - 180 degrees. */
		SHORT	screenId;		/* Screen ID 0 is reserved for process colours, else it's a spot colour. */
		SHORT	inkId;			/* The Id of the ink to use, for process colours, K = 0, C = 1, M = 2, Y = 3. */
		TCHAR	name[SZ_RENDERDATANAME]; /* For example "Process" */
		TCHAR	inkname[SZ_RENDERDATANAME]; /* For example "Black" */
	} RENDERDATA, *PRENDERDATA;
#endif
#endif
#include    <windows.h>
#include    <tchar.h>

	/// <summary>Some meta-information about interfaces.</summary>
	typedef struct _iInterface
	{
		/// <summary>Refers to the instance handle of the loaded library.</summary>
		HINSTANCE instance;
		/// <summary>One of INTERFACE_TYPE_xxxx.</summary>
		int type;
		/// <summary>An id, unique according to type above.</summary>
		int id;
		/// <summary>The short description is used to allow the user to select this device or interface out of a list box.</summary>
		LPCTSTR ShortDescription;
		/// <summary>The long description is used to allow the user to understand what this is all about.</summary>
		LPCTSTR LongDescription;
		/// <summary>Copyright information can be added here.</summary>
		LPCTSTR Copyright;
	} INTERFACETYPE, *PINTERFACETYPE;

/* NOTES:
The Interface id can be made up by using letters, the following are currently used

INTERFACE_TYPE_IMAGE:
'BMP' for bitmaps
'TIF' for TIFF images.

INTERFACE_TYPE_DIRECT:
'MXML' for metadata xml ouput
0 is used in the example DemoPlateDriver
*/

#ifdef __cplusplus
}

#ifdef _WIN32		//if compiling for windows
#include <pshpack2.h>	//set structure packing to 2 bytes!!!!
#endif
#ifdef __OS2__		//if compiling for os2
#pragma	pack(2)
#endif

typedef struct _margins
{
	///<summary>The left margin in 0.01 mm increments.</summary>
	unsigned left;
	///<summary>The top margin in 0.01 mm increments.</summary>
	unsigned top;
	///<summary>The right margin in 0.01 mm increments.</summary>
	unsigned right;
	///<summary>The bottom margin in 0.01 mm increments.</summary>
	unsigned bottom;
} MARGINS;

///<summary>
/// Data about a plate or film.
///</summary>
typedef struct _PlateData
{
	///<summary>Structure Identity</summary>
	STRCIDL strcIdl;
	///<summary>The full width in 0.01 mm increments. For the usable width, the margins are removed.</summary>
	unsigned width;
	///<summary>The full height in 0.01 mm increments. For the usable height, the margins are removed.</summary>
	unsigned height;
	///<summary>The plate margins.</summary>
	MARGINS margins;
	///<summary>The maximum number of pixels per inch supported. At 2400lpi and 16 x 16 dots, this would need to be at least 38400 ppi.</summary>
	unsigned maxResolution;
	///<summary>The human readable name of the plate, a zero terminated string of chars.</summary>
	char plateName[SZ_RENDERDATANAME];
} PLATEDATA, *LPPLATEDATA;

#define PLATEDATA_ID	'lP'	//'Pl' backwards

_inline void initPlateData(LPPLATEDATA lpPlate)
{
	lpPlate->strcIdl.cb = (USHORT)sizeof(PLATEDATA);
	lpPlate->strcIdl.ident = (USHORT)PLATEDATA_ID;
}

/* For interface type INTERFACE_METADATAXML only. */
typedef struct _metadataControl
{
	LPTSTR pFileName;		/* The metadata xml filename */
	ULONG flags;			/* One or both of the METACONTROL flags below */
	ULONG plateID;		    /* Index of plate */
	ULONG faceID;		    /* 0 = front face, 1 = back face */
	LPVOID pDriverdata;		/* Reserved as intermediate storage for the driver */
} METACONTROL, *LPMETACONTROL;

#define METACONTROL_OPENXML		0x01	/* The XML file should be opened when calling Open() */
#define METACONTROL_CLOSEXML	0x02	/* The XML file should be closed when calling Close() */

#ifdef _WIN32		//if compiling for windows
#include <poppack.h>
#endif
#ifdef __OS2__		//if compiling for os2
#pragma	pack()
#endif

///<summary>
/// This class defines the way we can perform the output of bits for a direct to plate or film device. 1 bit per pixel.
/// INTERFACE_PLATEPARALLEL: Adjacent output images will be output line for line across the plate, the resolution of the images must match the resolution on open.
/// ELSE: Images are output one at a time, after a SetPosition() command. Resolutions of images may differ.
///</summary>
typedef class DllExport _RasterOutputInterface
{
protected:
	ULONG	_operatingmode;
public:
	///<summary>
	/// Get data about a plate or film else null.
	/// This should work before open is called.
	///</summary>
	///<param name='index'>Input: Plate Data index, Output: total number of supported plate variants.</param>
	///<returns>Returns PLATEDATA or NULL.</returns>
	const LPPLATEDATA virtual EnumPlateData(int &index) { index = 0;  return NULL; }

	///<summary>
	///returns either INTERFACE_PLATEPARALLEL, INTERFACE_PAGEFORPAGE, or INTERFACE_METADATAXML, and optionally INTERFACE_AUTOPLATEFEED
	///</summary>
	ULONG OperatingMode() { return _operatingmode; }

	///<summary>
	/// Get a handle to a plate or film else INVALID_HANDLE_VALUE.
	///</summary>
	///<param name='pXmlMetadata'>For INTERFACE_METADATAXML only, else NULL.</param>
	///<param name='lpRd'>The data describing the ink, separation and other rendering intent information.</param>
	///<param name='PlateIndex'>The index points to the selected PLATEDATA (see EnumPlateData()) in this interface</param>
	///<returns>Returns a plate handle or NULL.</returns>
	HANDLE virtual Open(LPMETACONTROL pXmlMetadata, PRENDERDATA lpRd, int PlateIndex) { return INVALID_HANDLE_VALUE; }

	///<summary>
	/// Move to a position on the plate. For INTERFACE_PLATEPARALLEL, moves may be restricted in direction, the plate handle can be used to move (all pages are the same resolution).
	/// For INTERFACE_PAGEFORPAGE, or INTERFACE_METADATAXML, moves may be in any direction, the page handle is used to move (pages can have different resolutions).
	///</summary>
	///<param name='Handle'>The Handle created with Open() above, or with OpenRasterPage() below, depending on the operating mode.</param>
	///<param name='newPosition'>The new position from left and top margins in 0.01 mm increments if applicable, else NULL.</param>
	///<param name='positionFlags'>The positionFlags deterine what kind of positioning is desired.</param>
	///<returns>Returns zero on success, an error code on failure.</returns>
	int virtual SetPosition(HANDLE Handle, LPPOINT newPosition, ULONG positionFlags) = 0;

	///<summary>
	///Open a page to print on the plate, (mapsize.x / (dotmapSize * lpiResolution)) / 2540 is the width of the page in 0.01mm increments as 1 inch = 25.4mm.
	///The device may allow more than one page at a time, and it may expect to get sorted line for line data for all pages across the width of the plate.
	///Alternatively, the device can output each page individually to the plate. See OperatingMode()
	///</summary>
	///<param name='plateHandle'>The Handle created with Open() above.</param>
	///<param name='pDataFile'>INTERFACE_METADATAXML: The name of the raster image file to output, else NULL.</param>
	///<param name='plateOffset'>Offset on Plate from left and top margins in 0.01 mm increments to top and left of image.</param>
	///<param name='mapSize'>The width and height of the entire image in pixels.</param>
	///<param name='lpiResolution'>Cannot resolve to more than maxResolution of the selected PLATEDATA given the dotmapSize.</param>
	///<param name='dotmapSize'>the number of pixels in a raster dot from minimum 16 to maximum 32 (16x16 to 32x32).</param>
	///<returns>Returns a page handle or NULL. This handle most be usable to access the handle from Open() above.</returns>
	HANDLE virtual OpenRasterPage(HANDLE plateHandle, LPCTSTR pDataFile, SIZE plateOffset, SIZE mapSize, int lpiResolution, int dotmapSize) = 0;

	///<summary>
	///Close the handle to a Raster Page.
	///</summary>
	///<param name='pageHandle'>The handle which was created by OpenRasterPage.</param>
	///<param name='value'>Value to translate</param>
	///<param name='bGetMeasurement'>If true, get the pixel value in 0.01 mm increments, else the inverse.</param>
	///<returns>Returns the translated value for the given raster page handle.</returns>
	LONG virtual TranslateCoordinates(HANDLE pageHandle, LONG value, bool bGetMeasurement) = 0;

	///<summary>
	///Write data to the Rasteriser device, the data is in a format equivalent to a 1 bit per pixel bitmap.
	///</summary>
	///<param name='pageHandle'>The handle which was created by OpenRasterPage. Use this to figure out where the raster line starts.</param>
	///<param name='yRow'>Y Offset in image rows from 0 to mapSize.cy.</param>
	///<param name='lpBuffer'>The raster image buffer for a single image row of pixels in 1 bit per pixel bitmap format.</param>
	///<param name='rowSize'>The size of a single image row of pixels in bytes.</param>
	///<returns>Returns true on success, else false.</returns>
	bool virtual WriteRasterData(HANDLE pageHandle, LONG yRow, LPCVOID lpBuffer, size_t rowSize) { return true; };

	///<summary>
	///Close the handle to a Raster Page.
	///</summary>
	///<param name='pageHandle'>The handle which was created by OpenRasterPage.</param>
	void virtual CloseRasterPage(HANDLE pageHandle) = 0;

	///<summary>
	///Close the handle to a plate or film.
	///</summary>
	///<param name='plateHandle'>The handle which was created by OpenPlate.</param>
	void virtual Close(HANDLE plateHandle) = 0;

} RasterOutputInterface, *lpRasterOutputInterface;

///<summary>
/// This class defines the way we can perform the input of bits, mostly for a file format like tiff. 1 bit per pixel.
/// The class is designed for reading back rasterised data out of a data map, possibly in a non-linear fashion.
///</summary>
typedef class DllExport _RasterImageInterface
{
protected:
	ULONG	_orientation;
public:

	///<summary>
	///returns either INTERFACE_TOPDOWNIMG, or INTERFACE_BOTTOMUPIMG
	///</summary>
	ULONG Orientation() { return _orientation; }

	///<summary>
	///Checks if the file could be opened by this interface by verifying the file extension.
	///</summary>
	///<param name='lpFilePath'>The full name of the input or output file.</param>
	///<returns>Returns true if the file extension matches, else false.</returns>
	bool virtual TryFile(LPCTSTR lpFilePath) = 0;

	///<summary>
	///Open the read interface of the Image.
	///</summary>
	///<param name='lpFilePath'>The full name of the input or output file.</param>
	///<param name='bpp'>Output: Bits per pixel.</param>
	///<returns>Returns a handle or INVALID_HANDLE_VALUE on failure. A handle is of type void*.</returns>
	HANDLE virtual OpenRead(LPTSTR lpFilePath, int &bpp) = 0;

	///<summary>
	///Sets or changes the extension of a file to match the image type. Returns text to log the changes made.
	///</summary>
	///<param name='pFilePath'>The full name of the output file is a 260 byte buffer to write to, typically a changed intermediate file name is used.</param>
	///<returns>Returns a short file name, or whatever can be sensibly used for a log message.</returns>
	LPCTSTR virtual SetFileExtension(LPTSTR pFilePath) = 0;

	///<summary>
	///Open the write interface of the Image.
	///</summary>
	///<param name='lpFilePath'>The full name of the output file can be used to create a raster file in the same path.</param>
	///<param name='mapSize'>The width and height of the raster image in pixels.</param>
	///<param name='rgbMask'>One of the rgbMask values.</param>
	///<param name='bpp'>Input/Output: Bits per pixel, requested returns actual.</param>
	///<returns>Returns a handle or INVALID_HANDLE_VALUE on failure. A handle is of type void*.</returns>
	HANDLE virtual OpenWrite(LPTSTR lpFilePath, SIZE mapSize, LONG rgbMask, int &bpp) = 0;

	///<summary>
	///Reads the map information of the image.
	///</summary>
	///<param name='rowSize'>Returns the number of bytes in a row.</param>
	///<returns>Returns the size of the bitmap image.</returns>
	SIZE virtual GetMapSize(size_t &rowSize) = 0;

	///<summary>
	///Reads bits per pixel of the image.
	///</summary>
	///<returns>Returns the bits per pixel image.</returns>
	int virtual BitsPerPixel() = 0;

	///<summary>
	///Read image data.
	///</summary>
	///<param name='handle'>The handle which was created by OpenRasterInterface.</param>
	///<param name='offset'>The row at which to start reading data - will be updated on return.</param>
	///<param name='lpBuffer'>The raster image buffer in 1 bit per pixel bitmap format.</param>
	///<param name='nNumberOfBytesToRead'>The number of bytes to read is made up of one or more 32 bit words and must equal one or more rows of bitmap data.</param>
	///<returns>Returns true on success, else false.</returns>
	bool virtual ReadRasterData(HANDLE handle, PLONG offset, LPVOID lpBuffer, size_t nNumberOfBytesToRead) = 0;

	///<summary>
	///Write data to the image, the data is in a format equivalent to the pixel bitmap resolution.
	///</summary>
	///<param name='handle'>The handle which was created by OpenRasterInterface.</param>
	///<param name='rowSize'>The size of a single image row of pixels can be larger than the real number of pixels.</param>
	///<param name='lpBuffer'>The raster image buffer in 1 bit per pixel bitmap format.</param>
	///<param name='nNumberOfBytesToWrite'>The number of bytes to write will be a multiple of the rowSize.</param>
	///<returns>Returns true on success, else false.</returns>
	bool virtual WriteRasterData(HANDLE handle, size_t rowSize, LPCVOID lpBuffer, size_t nNumberOfBytesToWrite) = 0;

	///<summary>
	///Saves the raster image. This is called afer all data has been written. If this is not needed just return true for success.
	///</summary>
	///<param name='handle'>The handle which was created by OpenWrite.</param>
	///<returns>Returns true on success, else false.</returns>
	bool virtual Save(HANDLE handle) = 0;

	///<summary>
	///Close the interface of the raster image.
	///</summary>
	///<param name='handle'>The handle which was created by OpenRead or OpenWrite.</param>
	void virtual Close(HANDLE handle) = 0;
} RasterImageInterface, *lpRasterImageInterface;

extern "C" {
	typedef APIRET(APIENTRY _PFNRASTERINIT)(ULONG version, HMODULE mod, PULONG nInterfaces, PINTERFACETYPE* interfaces, PVOID mf);
	typedef _PFNRASTERINIT* PFNRASTERINIT;

	typedef PVOID(APIENTRY _PFNGETRASTEROP)(int type, int id);
	typedef _PFNGETRASTEROP* PFNGETRASTEROP;

/*******************************************************************************
	Instantiation of a class out of a dll, when you dont know what the class name really is.
	We use the most simple possible getter to solve this and cast to the base class defined above.

	And - we fix the name of the getter using a library definition file to the ID 10:
	Add a <name>.DEF file to the dll source
		LIBRARY <name>
		EXPORTS LibraryInit								@10
		        GetInterfaceInstance					@20
*******************************************************************************/

	///<summary>
	///Initialise the import library.
	///</summary>
	///<param name='version'>The Maul Publisher version number is hex 20000 + the version number of Maul Publisher (e.g. 340), that way a check for compatibility is possible.</param>
	///<param name='mod'>The module handle of self (the own module handle).</param>
	///<param name='nInterfaces'>The number of provided interfaces.</param>
	///<param name='interfaces'>A list of provided interfaces.</param>
	///<param name='mf'>The Maul Publisher API function list as PVOID. This can be used to output error messages for example, or it can just be ignored.</param>
	///<returns>Returns the version of the library.</returns>
	APIRET  APIENTRY LibraryInit(ULONG version, HINSTANCE mod, PULONG nInterfaces, PINTERFACETYPE* interfaces, PVOID mf);


	///<summary>
	///Create a new instance of the interface.
	///</summary>
	///<param name='type'>One of INTERFACE_TYPE_xxxx.</param>
	///<param name='id'>The id of the interface.</param>
	///<returns>Returns lpRasterOutputInterface or lpRasterImageInterface as PVOID on success, else NULL.</returns>
	PVOID APIENTRY GetInterfaceInstance(int type, int id);
}
#endif  /* __cplusplus */

#endif
#endif
